package Solartron1250;
#$Id: Solartron1250.pm,v 1.69 2022/01/18 11:52:20 sofc Exp $

############################################################
#
#    Package Solartron1250
#
# This module implements the FRA class for a Solartron
# --{\textregistered}
# 1250 frequency response analyser.
#
# It also implements the Solartron1250SL class in order to properly handle
# the different versions of the 1250 (the SL class being used for the version 
# of the 1250 supplied with the Slumberger logo
#
# To get a device instance of one of the classes call the constructors:
#
# $fra = Solartron1250->new($gpib_address);
#
# $fra = Solartron1250SL->new($gpib_address);
#
############################################################ 


use strict;
use warnings;
use FDEV;
our @ISA = qw(FRA);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Solartron 1250';
    $self->{'_virtual'} = 0;
    $self->{'_mode'} = 'V';
    $self->{'_modes'} = {'Volt' => 'V'},
    $self->{'_minf'} = 1E-5;
    $self->{'_maxf'} = 65000;
    $self->{'_minpV'} = -10.22;
    $self->{'_maxpV'} = 10.22;
    $self->{'_minpA'} = undef;
    $self->{'_maxpA'} = undef;
    $self->{'_maxampA'} = undef;
    $self->{'_minampA'} = undef;
    $self->{'_maxampV'} = 10.23;
    $self->{'_source_modes'} = {'V1/V2' => '0102', 'V2/V1' => '0201','V1' => '0100', 'V2' => '0200'};
    $self->{'_errorcodes'} = {
        '00' => ' ',
        '01' => 'Unknown command',
        '02' => 'Argument mismatch',
        '03' => 'Argument out of range',
        '04' => 'Floating point format error',
        '05' => 'Illegal request for value',
	'30' => 'Generator stopped',
	'31' => 'Generator overload',
	'32' => 'Generator stopped by external contact closure',
        'XX' => 'No data recieved',
        '80' => ' ',
        '81' => 'Input Overload',
        '82' => 'Auto integration failed',
        '83' => 'Input Overload and Autointegration failed',
        '84' => 'Amplitude compression failed',
        '85' => 'Input Overload and Amplitude compression failed',
        '86' => 'Autointegration failed and Amplitude compression failed',
        '87' => 'Input overload, Amplitude compression failed and Autointegration failed',
        '88' => 'Internal hardware error',
        '89' => 'Analysis terminated early',
    };
    $self->{'_overhead'} = 1.05;
    $self->{'_minintegration'} = 0.1;
    $self->{'_setuptime'} = 15;
    $self->{'_delay'} = 0.1;
    $self->{'_frequency_overhead_time'} = 6;
    $self->{'_setup_tags'} = 'mode,input_mode_V1,input_mode_V2,measuring_range_V1,measuring_range_V2,polarisation,amplitude,input_V1,input_V2,polarisation_offset,source_mode,input_coupling_V1,input_coupling_V2,delay,harmonic';
    $self->{'_default'} = {
        'mode' => 'Volt',
        'input_mode_V1' => 'Differential',
        'input_mode_V2' => 'Differential',
        'source_mode' => 'V2/V1',
        'input_V1' => 'Rear',
        'input_V2' => 'Rear',
        'input_coupling_V1' => 'DC',
        'input_coupling_V2' => 'DC',
        'measuring_range_V1' => 'auto',
        'measuring_range_V2' => 'auto',
        'polarisation_offset' => 'OFF',
        'reference_potential' => 'Fixed',
	'harmonic' => '1',
        'delay' => $self->{'_delay'},
        'amplitude' => $self->{'_minampV'},
        'polarisation' => 0,
    };
    $self->{'_desc'} = {
        'mode' => 'Control mode',
        'input_mode_V1' => 'Input mode for V1 input, either differential or single ended',
        'input_mode_V2' => 'Input mode for V2 input, either differential or single ended',
	'input_V1' => 'Input terminal for V1',
        'input_V2' => 'Input terminal for V2',
	'harmonic' => 'Determines if the base frequency or one of the higher harmonics are used for measuring impedance',
        'source_mode' => 'Determines how the impedance is calculated',
        'input_coupling_V1' => 'Determines if the input coupling for voltage input 1 is direct through or uses a high pass filter',
        'input_coupling_V2' => 'Determines if the input coupling for voltage input 2 is direct through or uses a high pass filter',
        'measuring_range_V1' => 'Input range for voltage input 1, either fixed or autorange',
        'measuring_range_V2' => 'Input range for voltage input 2, either fixed or autorange',
        'polarisation_offset_program' => 'Select if a polarisation offset program (and which one) is to be run prior to impedance scans',
        'reference_potential' => 'Select if the polarisation is to be set versus chassis ground or versus measured potential at OCV',
        'delay' => 'Delay in seconds between each frequency',
        'amplitude' => 'AC amplitude (RMS, V or mA depending on mode)',
        'polarisation' => 'DC polarisation (V or mA depending on mode)',
    };
    $self->{'_values_harmonic'} = [1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16];
    $self->{'_values_source_mode'} = ['V1' , 'V2' , 'V1/V2', 'V2/V1'];
    $self->{'_values_input_coupling_V1'} = ['AC','DC'];
    $self->{'_values_input_coupling_V2'} = ['AC','DC'];
    $self->{'_values_measuring_range_V1'} = ['auto',0.03,0.3,3,30,300];
    $self->{'_values_measuring_range_V2'} = ['auto',0.03,0.3,3,30,300];
    $self->{'_values_input_V1'} = ['Front','Rear'];
    $self->{'_values_input_V2'} = ['Front','Rear'];
    $self->{'_save_tags'} = ['FR','IS','AM','BI','MS','RA1','RA2','DC1','DC2','IP1','IP2','SO','CO','HA'];
    bless $self,$class;
    return $self;
}

sub save()
{
    my $self = shift;
    my @res = ();
    foreach my $tag (@{$self->{'_save_tags'}})
    {
        $self->writecmd('?'.$tag);
        my $r = $self->readcmd();
        if ($r =~ /([\d\.\+\-eE\,]+)/)
        {
            push @res,$tag.' '.$1;
        }
    }
    return @res;
}

sub bulk_load(@)
{
    my $self = shift;
    $self->writecmd('OT 3'); # Fast reply cr + EOI #
    sleep(1);
    $self->writecmd('OT 1');
    $self->writecmd('OP 2,1');  # GPIB does all #
    while (@_)
    {
        my $cmd = shift;
        $self->checkcmd($cmd);
    }
    return undef;
}

sub init()
{
    my $self = shift;
    $self->{'_pol_offset'} = 0;
    $self->writecmd('TT2');
    sleep(5);
    $self->writecmd('OT 3');
    sleep(1);
    ### Do 'OT 1' twice as first time generates error.... ###
    ### (For who knows what reason, manual says nothing   ###
    ### of this...)                                       ###
    $self->writecmd('OT 1');
    sleep(1);
    $self->writecmd('OT 1');
    $self->writecmd('OP 2,1');
    $self->checkcmd(sprintf("AM %.4g",$self->{'_minampV'}));
    $self->writecmd('BI 0.000');
    return 'Default';
}

sub isallive()
{
    my $self = shift;
    $self->writecmd('OT 1');
    sleep(1);
    $self->writecmd('OT 1');
    $self->writecmd('OP 2,1');
    $self->writecmd('?TS');
    my $res = $self->readcmd();
    return 1 if (defined($res) && $res =~ /\d/);
    return 0;
}

sub reset()
{
    my $self = shift;
    $self->writecmd('SG');
    $self->writecmd('BK');
    sleep(1);
    $self->writecmd('SG');
    $self->writecmd('BK');
    sleep(1);
}

sub set_polarisation($$)
{
    my $self = shift;
    my $pol = shift;
    my $setup = shift;
    my $ext_pol = $self->get_setup_value($self->get_default('polarisation_offset'),$setup,'polarisation_offset');
    if (ref($ext_pol) eq 'RemoteExec')
    {
	$self->writedebug("Executeing function to get polarisation offset");
        $self->{'_pol_offset'} = $ext_pol->exec_id($self);
        $pol += $self->{'_pol_offset'};
    }
    $pol = $self->{'_minpV'} if ($pol < $self->{'_minpV'});
    $pol = $self->{'_maxpV'} if ($pol > $self->{'_maxpV'});
    $self->checkcmd(sprintf("BI %.4g",$pol));
    return $pol;
}

sub amplitude
{
    my $self = shift;
    return $self->{'_amplitude'} unless (@_);
    my $a = shift;
    $a = $self->{'_minampV'} if ($a < $self->{'_minampV'});
    $a = $self->{'_maxampV'} if ($a > $self->{'_maxampV'});
    $self->checkcmd(sprintf("AM %.4g",$a));
    $self->{'_amplitude'} = $a;	
    return $self->{'_amplitude'};
}

sub setup_impedance($)
{
    my $self = shift;
    my $setup = shift;
    $self->init();
    $self->checkcmd('WF 0');  ### Sinus waweform ###
    $self->checkcmd('MO 0');  ### Modulation off ###
    $self->checkcmd('SQ 0');  ### stop at 0 deg  ###
    $self->checkcmd('AU 0');  ### autointegration off (we controll all) ###
    $self->{'_delay'} = $self->get_setup_value($self->{'_delay'},$setup,'delay');
    $self->checkcmd(sprintf("MS %.4g",$self->{'_delay'}));
    ### Measure range on input ###
    my $ra1 = $self->get_setup_value($self->get_default('measuring_range_V1'),$setup,'measuring_range_V1');
    if ($ra1)
    {
        my $tr = 0; ## Default autorange ##
        my $x = 0;
        foreach (@{$self->{'_values_measuring_range_V1'}})
        {
            $tr = $x if ($_ eq $ra1);
            $x++
        }
        $ra1 = $tr;
    }
    my $ra2 = $self->get_setup_value($self->get_default('measuring_range_V2'),$setup,'measuring_range_V2');
    if ($ra2)
    {
        my $tr = 0; ## Default autorange ##
        my $x = 0;
        foreach (@{$self->{'_values_measuring_range_V2'}})
        {
            $tr = $x if ($_ eq $ra2);
            $x++
        }
        $ra2 = $tr;
    }
    $self->checkcmd('RA 1,'.$ra1);
    $self->checkcmd('RA 2,'.$ra2);
    ### DC coupling default ###
    my $ac1 = 0;
    $ac1 = 1 if ($self->get_setup_value($self->get_default('input_coupling_V1'),$setup,'input_coupling_V1') =~ /AC/i);
    my $ac2 = 0;
    $ac2 = 1 if ($self->get_setup_value($self->get_default('input_coupling_V2'),$setup,'input_coupling_V2') =~ /AC/i);
    $self->checkcmd('DC 1,'.$ac1);
    $self->checkcmd('DC 2,'.$ac2);
    ### Input front/rear, default rear ###
    my $i1 = 1;
    my $i2 = 1;
    $i1 = 0 if ($self->get_setup_value($self->get_default('input_V1'),$setup,'input_V1') =~ /F/i);
    $i2 = 0 if ($self->get_setup_value($self->get_default('input_V2'),$setup,'input_V2') =~ /F/i);
    $self->checkcmd("IP 1,$i1");
    $self->checkcmd("IP 2,$i2");
    my $source_mode = $self->get_setup_value($self->get_default('cource_mode'),$setup,'source_mode');
    ### Default V2/V1  ##
    my $sm = '0201';
    $sm = $self->{'_source_modes'}{$source_mode} if exists($self->{'_source_modes'}{$source_mode});
    $self->checkcmd('SO '.$sm);
    $self->writecmd('CO 0');  ### cordinates set to a,b ###
    ### Set polarisation and amplitude ###
    my $pol = $self->get_setup_value(0,$setup,'polarisation');
    my $truepol = $self->set_polarisation($pol,$setup);
    $self->amplitude($self->get_setup_value(undef,$setup,'amplitude'));
    $self->writecmd('RG');
    sleep(2);
    $self->reset_data;
    return $truepol;
}


sub run_frequency($$)
{
    my $self = shift;
    my $freq = shift;
    my $time = shift;
    my $setup = shift;
    my $testfile = undef;
    $testfile = $$setup{'semafore_file'} if (exists $$setup{'semafore_file'} && $$setup{'semafore_file'});
    return undef unless (!$testfile || -e $testfile);
    my $h = int($self->get_setup_value(1,$setup,'harmonic'));
    $self->checkcmd('HA '.$h);
    $freq = $self->{'_minf'} if ($freq < $self->{'_minf'});
    $freq = $self->{'_maxf'}/$h if ($freq > $self->{'_maxf'}/$h);
    $time = $self->get_integration_time($freq,$time);
    my $fout = sprintf("%.4g",$freq);
    my $tout = sprintf("%.4g",$time);
    $self->checkcmd('FR '.$fout);
    $self->checkcmd('IS '.$tout);
    $self->writecmd('SI');  ## Run impedance ###
    my $etime = time + $time;
    my $maxtime = $etime + $time;
    my $f = 1;
    $f = $$setup{'resistor'} if (exists($$setup{'resistor'}));
    my $count = 1;
    while (time < $etime)
    {
        sleep(1);
	if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->() if (ref($$setup{'function'}) eq 'CODE');
        }
        next if (!$testfile || -e $testfile);
	$self->reset();
	push @{$self->{'_data'}},undef;
	return undef;
    }
  OUTER:
    while (time < $maxtime)
    {
        my $readloop = 0;
      INNER:
        while ($readloop < 5)
        {
	    $self->writecmd('OT 1');
	    $self->writecmd('OP 2,1');
            my $res = $self->readcmd();
            if (!defined($res) || $res =~ /error/i || $res !~ /\w/)
            {
		last unless (!$testfile || -e $testfile);
		my $t = $time;
		$t = 10 if ($t > 10);
		sleep($t);
                $readloop++;
                next INNER;
            }
	    unless ($res)
	    {
		$res = '0.0000,0.0000,0.0000';
		$self->writedebug('No response from 1250, using 0,0,0 instead!');
	    }
            my @r = split(",",$res);
            my $result = sprintf("%f %f %f",$r[0],$r[1]*$f,$r[2]*$f);
	    push @{$self->{'_data'}},$result;
            $self->{'_last_error'} = $r[3];
            if ($r[3] =~ /(\d+)/)
            {
                $self->{'_last_error'} = '8'.$1;
            }
	    ### 1250 does not report any warning codes ###
	    return $result;
        }
        sleep(1);
	if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->() if (ref($$setup{'function'}) eq 'CODE');
        }
        next OUTER unless (!$testfile || -e $testfile);
	$self->reset();
	push @{$self->{'_data'}},undef;
        $self->{'_last_error'} = 'XX';
	return undef;
    }
    $self->writecmd('SG');
    push @{$self->{'_data'}},undef;
    $self->{'_last_error'} = 'XX';
    return undef;
}


sub checkcmd($)
{
    ########################################################
    #### checkcmd($)                                     ###
    #### Simple wrapper for writecmd() as the 1250 has   ###
    #### no CE command and thus no way of using the      ###
    #### ?ER command for anything usefull.....           ###
    ########################################################
    my $self = shift;
    my $cmd = shift;
    $self->writecmd($cmd);
}

sub checkerror()
{
    my $self = shift;
    $self->writecmd('?ER');
    my $rep = $self->readcmd();
    if (!$rep || $rep !~ /^00/)
    {
        $self->errorlog("ERROR cummunicating with $self->{'_type'}: message: '$rep'");
        if ($rep = /~(\d+)/)
        {
            $self->errorlog('Error description: '.$self->errorcode($1));
        }
    }
}

sub get_error()
{
    my $self = shift;
    $self->writecmd('?ER');
    my $rep = $self->readcmd();
    if ($rep =~ /(\d+)/)
    {
        my $r = $self->errorcode($1);
	$self->writedebug("Error description: $r");
	return $r;
    }
    return '';
}



package Solartron1250SL;
our @ISA = qw(Solartron1250);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Solartron 1250 SL';
    bless $self,$class;
    return $self;
}

sub run_frequency($$)
{
    my $self = shift;
    my $freq = shift;
    my $time = shift;
    my $setup = shift;
    my $testfile = undef;
    $testfile = $$setup{'semafore_file'} if (exists $$setup{'semafore_file'} && $$setup{'semafore_file'});
    return undef unless (!$testfile || -e $testfile);
    my $h = int($self->get_setup_value(1,$setup,'harmonic'));
    $self->checkcmd('HA '.$h);
    $freq = $self->{'_minf'} if ($freq < $self->{'_minf'});
    $freq = $self->{'_maxf'}/$h if ($freq > $self->{'_maxf'}/$h);
    $time = $self->get_integration_time($freq,$time);
    my $fout = sprintf("%.4g",$freq);
    my $tout = sprintf("%.4g",$time);
    $self->checkcmd('FR '.$fout);
    $self->checkcmd('RG');
    $self->checkcmd('IS '.$tout);
    $self->writecmd('SI');  ## Run impedance ###
    ### Default V2/V1  ##                                                                                       
    my $sm = '0201';
    my $source_mode = $self->get_setup_value($self->get_default('cource_mode'),$setup,'source_mode');
    $sm = $self->{'_source_modes'}{$source_mode} if exists($self->{'_source_modes'}{$source_mode});
    my $etime = time + $time;
    my $maxtime = $etime + $time;
    my $f = 1;
    $f = $$setup{'resistor'} if (exists($$setup{'resistor'}));
    my $count = 1;
    while (time < $etime)
    {
        sleep(1);
	if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->() if (ref($$setup{'function'}) eq 'CODE');
        }
        next if (!$testfile || -e $testfile);
	$self->reset();
	push @{$self->{'_data'}},undef;
	return undef;
    }
  OUTER:
    while (time < $maxtime)
    {
        my $readloop = 0;
      INNER:
        while ($readloop < 5)
        {
	    $self->writecmd('SO '.$sm);
#	    $self->writecmd('OT 1');
#	    $self->writecmd('OP 2,1');
            my $res = $self->readcmd();
            if (!defined($res) || $res =~ /error/i || $res !~ /\w/)
            {
		last unless (!$testfile || -e $testfile);
		my $t = $time;
		$t = 10 if ($t > 10);
		sleep($t);
                $readloop++;
                next INNER;
            }
	    unless ($res)
	    {
		$res = '0.0000,0.0000,0.0000';
		$self->writedebug('No response from 1250, using 0,0,0 instead!');
	    }
            my @r = split(",",$res);
            my $result = sprintf("%f %f %f",$r[0],$r[1]*$f,$r[2]*$f);
	    push @{$self->{'_data'}},$result;
            $self->{'_last_error'} = $r[3];
            if ($r[3] =~ /(\d+)/)
            {
                $self->{'_last_error'} = '8'.$1;
            }
	    ### 1250 does not report any warning codes ###
	    return $result;
        }
        sleep(1);
	if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->() if (ref($$setup{'function'}) eq 'CODE');
        }
        next OUTER unless (!$testfile || -e $testfile);
	$self->reset();
	push @{$self->{'_data'}},undef;
        $self->{'_last_error'} = 'XX';
	return undef;
    }
    $self->writecmd('SG');
    push @{$self->{'_data'}},undef;
    $self->{'_last_error'} = 'XX';
    return undef;
}



sub run_frequency_all($$)
{
    my $self = shift;
    my $result = $self->run_frequency(@_);
    $result .= ' #';
    $self->writecmd('?SO');
    my $r = $self->readcmd();
    my $org = undef;
    if ($r =~ /([\d\,]+)/)
    {
        $org = 'SO '.$1;
    }
    foreach my $m (@{$self->{'_values_source_mode'}})
    {
        $self->writecmd('SO '.$self->{'_source_modes'}{$m});
#	$self->writecmd('OT 1');
#	$self->writecmd('OP 2,1');
        my $res = $self->readcmd();
	my $count = 5;
	while (!$res && $count--)
	{
	    ### Try a few times more ###
	    $self->writecmd('SO '.$self->{'_source_modes'}{$m});
#	    $self->writecmd('OT 1');
#	    $self->writecmd('OP 2,1');
	    $res = $self->readcmd();	    
	}
	### Supply dummy data if 5 tries didn't work ###
	unless ($res)
	{
	    $res = '0.0000,0.0000,0.0000';
	    $self->writedebug('No response from 1250, using 0,0,0 instead!');
	}
        my @r = split(",",$res);
        $result .= sprintf(" $m %f %f %f",$r[0],$r[1],$r[2]);
    }
    $result .= ' #';
    foreach my $c (1,2)
    {
	$self->writecmd('?RA'.$c);
	my $res = $self->readcmd();
	unless (defined($res))
	{
	    ### try once more if first try did not work ###
	    $self->writecmd('?RA'.$c);
	    $res = $self->readcmd();
	}
        if (defined($res) && $res =~ /(\d+)/)
        {
            $res = $1+0;
        }
        if ($res)
        {
            $result .= " Range V$c: ".$self->{'_values_measuring_range_V'.$c}[$res];
        }
        else
        {
            $self->writecmd('?AR'.($c-1));
            my $res = $self->readcmd();
            if (defined($res) && $res =~ /(\d+)/)
            {
                $res = $1+0;
            }
            $result .= " Range V$c: ".$self->{'_values_measuring_range_V'.$c}[$res];
        }
	$self->writecmd('?ER');
	$self->readcmd();
    }
#    $self->writecmd($org);
#    $self->writecmd('OP 2,1');
    push @{$self->{'_longdata'}},$result;
    return $result;
}


sub reset()
{
    my $self = shift;
    $self->writecmd('BK');
    sleep(1);
    $self->writecmd('BK');
    sleep(1);
}

return 1;
