package Solartron1280;

############################################################
#
#    Package Solartron1280
#
# This module defines a number of classes for using a Solartron
# --{\textregistered}
# frequency response analyser/potentiostat. As the 1280 acts as a conbination of
# a 1250 and a 1287 the following classes are defined: Solartro1280/Solartron1280A 
# for the potentiostatic part of the device and Solartron1280_FRA for the 
# frequency response analyser part of the device. In order to get a device
# instance of either of the classes, call one of the constructors:
#
# $fra = Solartron1280_FRA->new($gpib_address);
#
# $pstat = Solartron1280->new($gpib_address);
#
# $pstat = Solartron1280A->new($gpib_address);
#
############################################################

use Solartron1287;
use Solartron1250;

our @ISA = qw(Solartron1287);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Solartron 1280';
    $self->{'_mode'} = 'V';
    $self->{'_minf'} = 1E-3;
    $self->{'_maxf'} = 20000;
    $self->{'_ranges'} = [0,0.1,2,20,200];
    $self->{'_values_range'} = ['auto',0.1,1,2,20,200];
    $self->{'_save_tags'} = ['RG','FI','BY','ON','SY','PB','GB','RR','PO','PV','PC','DL','OL','GP','TR','BY','SW','IL','PX','PY','PI','VT','IT','BR','PW','OF'];
    bless $self,$class;
    return $self;
}

package Solartron1280A;
our @ISA = qw(Solartron1280);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Solartron 1280A';
    $self->{'_minpV'} = -12.8;
    $self->{'_maxpV'} = 12.8;
    $self->{'_minpA'} = -1;
    $self->{'_maxpA'} = 1;
    $self->{'_values_measure_resistor'} = ['auto',0.1,1,10,100,1000,10000,100000];
    $self->{'_resistors'} = [0,0.1,1,10,100,1000,10000,100000];
    bless $self,$class;
    return $self;
}


package Solartron1280_FRA;

our @ISA = qw(Solartron1250);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Solartron 1280 FRA';
    $self->{'_minf'} = 1E-3;
    $self->{'_maxf'} = 20000;
    $self->{'_minpV'} = -10;
    $self->{'_maxpV'} = 10;
    $self->{'_mode'} = 'V';
    $self->{'_modes'} = {'Volt' => 'V'},
    $self->{'_minpA'} = undef;
    $self->{'_maxpA'} = undef;
    $self->{'_maxampA'} = undef;
    $self->{'_minampA'} = undef;
    $self->{'_maxampV'} = 7;
    $self->{'_source_modes'} = { 'V2/V1' => '0201','V1' => '0100', 'V2' => '0200'};
    $self->{'_errorcodes'} = {
	'00' => ' ',
	'01' => 'Unknown command',
	'02' => 'Argument mismatch',
	'03' => 'Argument out of range',
	'04' => 'Floating point format error',
	'05' => 'Illegal request for value',
        '30' => 'Generator stopped',
        'XX' => 'No data recieved',
        '80' => ' ',
        '81' => 'Input Overload',
	'0' => ' ',
	'1' => 'Input Overload',
    };
    $self->{'_overhead'} = 1.05;
    $self->{'_minintegration'} = 1;
    $self->{'_delay'} = 0;  ## No delay possible ###
    $self->{'_setup_tags'} = 'measuring_range_V1,measuring_range_V2,polarisation,amplitude,polarisation_offset,harmonic,source_mode';
    $self->{'_default'} = {
        'mode' => 'Volt',
        'source_mode' => 'V2/V1',
        'measuring_range_V1' => 'auto',
        'measuring_range_V2' => 'auto',
        'polarisation_offset' => 'OFF',
        'reference_potential' => 'Fixed',
        'harmonic' => '1',
	'polarisation' => 0,
	'delay' => $self->{'_delay'},
	'ampitude' => $self->{'_minampV'},
    };
    $self->{'_values_source_mode'} = ['V1' , 'V2' , 'V2/V1'];
    $self->{'_values_harmonic'} = [1,2,3,4,5,6,7,8];
    $self->{'_values_measuring_range_V1'} = ['auto',0.03,0.3,3,30,300];
    $self->{'_values_measuring_range_V2'} = ['auto',0.03,0.3,3,30,300];
    $self->{'_save_tags'} = ['FR','IS','AM','BI','RA1','RA2','SO','CO','HA'];
    bless $self,$class;
    return $self;
};

sub init()
{
    my $self = shift;
    $self->{'_pol_offset'} = 0;
    $self->writecmd('TT1');
    sleep(5);
    $self->writecmd('OT 3');
    sleep(1);
    ### Do 'OT 1' twice as first time generates error.... ###
    ### (For who knows what reason, manual says nothing   ###
    ### of this...)                                       ###
    $self->writecmd('OT 1');
    sleep(1);
    $self->writecmd('OT 1');
    $self->writecmd('OP 2,1');
    $self->writecmd('AM '.sprintf("%.4g",$self->{'_minampV'}));
    return 'Default';
}

sub isallive()
{
    my $self = shift;
    $self->writecmd('OT 3');
    sleep(1);
    ### Do 'OT 1' twice as first time generates error.... ###
    ### (For who knows what reason, manual says nothing   ###
    ### of this...)                                       ###
    $self->writecmd('OT 1');
    sleep(1);
    $self->writecmd('OT 1');
    $self->writecmd('OP 2,1');
    $self->writecmd('?TS0');
    my $e = $self->readcmd();
    return 1 if (defined($e) && $e =~ /\d/);
    return 0;
}

sub run_frequency($$)
{
    my $self = shift;
    my $freq = shift;
    my $time = shift;
    my $setup = shift;
    my $testfile = undef;
    $testfile = $$setup{'semafore_file'} if (exists $$setup{'semafore_file'} && $$setup{'semafore_file'});
    return undef unless (!$testfile || -e $testfile);
    my $h = int($self->get_setup_value(1,$setup,'harmonic'));
    $self->checkcmd('HA '.$h);
    $freq = $self->{'_minf'} if ($freq < $self->{'_minf'});
    $freq = $self->{'_maxf'}/$h if ($freq > $self->{'_maxf'}/$h);
    $time = $self->get_integration_time($freq,$time);
    my $fout = sprintf("%.4g",$freq);
    my $tout = sprintf("%.4g",$time);
    $self->checkcmd('FR '.$fout);
    $self->checkcmd('IS '.$tout);
    $self->writecmd('SI');  ## Run impedance ###
    my $etime = time + $time;
    my $maxtime = $etime + $time;
    my $f = 1;
    $f = $$setup{'resistor'} if (exists($$setup{'resistor'}));
    my $count = 1;
    while (time < $etime)
    {
        sleep(1);
        if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->() if (ref($$setup{'function'}) eq 'CODE');
        }
        next if (!$testfile || -e $testfile);
        $self->reset();
        push @{$self->{'_data'}},undef;
        return undef;
    }
  OUTER:
    while (time < $maxtime)
    {
        my $readloop = 0;
      INNER:
        while ($readloop < 5)
        {
            $self->writecmd('OT 1');
            $self->writecmd('OP 2,1');
	    $self->writecmd('?NR');
            my $res = $self->readcmd();
            if ($res =~ /error/i)
            {
                $readloop++;
                next INNER;
            }
	    $self->writecmd(sprintf("UF%d",$res));
            $res = $self->readcmd();
            if ($res =~ /error/i)
            {
                $readloop++;
                next INNER;
            }
            my @r = split(",",$res);
            my $result = sprintf("%f %f %f",$r[0],$r[1]*$f,$r[2]*$f);
            push @{$self->{'_data'}},$result;
            $self->{'_last_error'} = $r[3];
            if ($r[3] =~ /(\d+)/)
            {
                $self->{'_last_error'} = '8'.$1;
            }
            ### 1250 does not report any warning codes ###
            return $result;
        }
        sleep(1);
        if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->() if (ref($$setup{'function'}) eq 'CODE');
        }
        next OUTER if (!$testfile || -e $testfile);
        $self->reset();
        push @{$self->{'_data'}},undef;
        $self->{'_last_error'} = 'XX';
        return undef;
    }
    $self->writecmd('SG');
    push @{$self->{'_data'}},undef;
    $self->{'_last_error'} = 'XX';
    return undef;
}

sub run_frequency_all($$)
{
    ###############################################################
    ### run_frequency_all($$)                                   ###
    ### runs a single frequency measurement as run_frequency()  ###
    ### unlike the normal function, this function returns as    ###
    ### many data as possible (including the valtages of the    ###
    ### individual inputs if possible)                          ###
    ### Defaults to run_frequency($$) if the device does not    ###
    ### support the additional functionality nescesarry         ###
    ###############################################################
    my $self = shift;
    my $string = $self->run_frequency(@_);
    push @{$self->{'_longdata'}},$string;
    return $string;
}


sub setup_impedance($)
{
    my $self = shift;
    my $setup = shift;
    $self->init();
    $self->reset();
    $self->checkcmd('WV 0');  ### Sinus waweform ###
    $self->checkcmd('SQ 0');  ### stop at 0 deg  ###
    $self->checkcmd('FN 0');  ### Scaling off  ###
    ### Measure range on input ###
    my $ra1 = $self->get_setup_value($self->get_default('measuring_range_V1'),$setup,'measuring_range_V1');
    if ($ra1)
    {
        my $tr = 0; ## Default autorange ##
        my $x = 0;
        foreach (@{$self->{'_values_measuring_range_V1'}})
        {
            $tr = $x if ($_ eq $ra1);
            $x++
        }
        $ra1 = $tr;
    }
    my $ra2 = $self->get_setup_value($self->get_default('measuring_range_V2'),$setup,'measuring_range_V2');
    if ($ra2)
    {
        my $tr = 0; ## Default autorange ##
        my $x = 0;
        foreach (@{$self->{'_values_measuring_range_V2'}})
        {
            $tr = $x if ($_ eq $ra2);
            $x++
        }
        $ra2 = $tr;
    }
    $self->checkcmd('RA 1,'.$ra1);
    $self->checkcmd('RA 2,'.$ra2);
    my $source_mode = $self->get_setup_value($self->get_default('source_mode'),$setup,'source_mode');
    ### Default V2/V1  ##
    my $sm = '0201';
    $sm = $self->{'_source_modes'}{$source_mode} if exists($self->{'_source_modes'}{$source_mode});
    $self->checkcmd('SO '.$sm);
    $self->writecmd('CO 0');  ### cordinates set to a,b ###
    ### Set polarisation and amplitude ###
    my $pol = $self->get_setup_value(0,$setup,'polarisation');
    my $truepol = $self->set_polarisation($pol);
    my $a = $self->get_setup_value(undef,$setup,'amplitude');
    if ($a)
    {
        $a = $self->{'_minampV'} if ($a < $self->{'_minampV'});
        $a = $self->{'_maxampV'} if ($a > $self->{'_maxampV'});
        $self->checkcmd('AM '.sprintf("%.4g",$a));
    }
    return $truepol;
}


sub checkcmd($)
{
    ########################################################
    #### checkcmd($)                                     ###
    #### runs command twice if first go results in error ###
    ########################################################
    my $self = shift;
    my $cmd = shift;
    $self->writecmd($cmd);
    $self->writecmd('?ER');
    my $e = $self->readcmd();
    if (!$e || $e !~ /^00/)
    {
        if ($e = /~(\d+)/)
        {
            $self->writedebug('Error description: '.$self->errorcode($1));
        }
        $self->writecmd('CE');
        $self->writecmd($cmd);
        $self->checkerror();
    }
}

sub checkerror()
{
    my $self = shift;
    $self->writecmd('?ER');
    my $rep = $self->readcmd();
    if (!$rep || $rep !~ /^00/)
    {
        $self->errorlog("ERROR cummunicating with $self->{'_type'}: message: '$rep'");
        if ($rep = /~(\d+)/)
        {
            $self->errorlog('Error description: '.$self->errorcode($1));
        }
        $self->writecmd('CE');
    }
}

return 1;
